#ifndef _HISTOGRAM_OP_
#define _HISTOGRAM_OP_

#include <vector>
#include "Image.h"



/*
 * This class is an implementation of several histogram related operations/queries.
 * Mainly used to normalize images to a given interval [minVal,maxVal] usually this
 * is [0,255].
 *
 * NOTE: This class uses the Standard Template Library (STL).
 *                      
 * Author: Ziv Yaniv (zivy@cs.huji.ac.il)
 */

using namespace std;

class HistogramOp {
public:
	static const float PRECISION;

	/*
	 * Returns the heighest gray level value in the image. 
	 */
	static float maxGrayValue(Image &img);
	static float maxGrayValue(vector<Image *> &imgs);

	/*
	 * Returns the smallest gray level value in the image. 
	 */
	static float minGrayValue(Image &img);
	static float minGrayValue(vector<Image *> &imgs);

	/*
	 * Returns the gray level value with the minimum number of appearances in the 
	 * image. If there are multiple results returns the first one found.
	 */
	static float min(Image &img);
	static float min(vector<Image *> &imgs);

	/*
	 * Returns the gray level value with the maximum number of appearances in the 
	 * image. If there are multiple results returns the first one found.
	 */
	static float mode(Image &img);
	static float mode(vector<Image *> &imgs);

	/*
	 * Normalize the given image so it has values in the interval [minVal,maxVal].
	 * A value is mapped as follows: 
	 * Given that the original image had values in the interval [min,max]
	 *                               val - min 
	 *             val -->  minVal + --------- *(maxVal - minVal)
	 *                               max - min
	 */
	static void normalize(Image &img, float minVal=0.0f, float maxVal=255.0f) {
		normalizeClamp(img,0.0,0.0,minVal,maxVal);
	}
	static void normalize(vector<Image *> &imgs, float minVal=0.0f, float maxVal=255.0f) {
		normalizeClamp(imgs,0.0,0.0,minVal,maxVal);
	}

	/*
	 * Normalize the given image so it has values in the interval [minVal,maxVal].
	 * A value is mapped as follows: 
	 * Given that the original image had values in the interval [min,max]
	 *                    __
	 *                    | minVal                                    if val<lowerThresh                            
	 *                    |  
	 *                    |          val - min 
	 *             val -->| minVal + --------- *(maxVal - minVal)     if lowerThresh<val<upperThresh
	 *                    |          max - min
   *                    | 
	 *                    | maxVal                                    if val>upperThresh
	 *                    -- 
	 * where lowerThresh is the gray level value which lowerThresholdPercentage of the image
	 * pixels are below and upperThresh is the gray level value lowerThresholdPercentage of the
	 * image pixels are above.
	 * 
	 * lowerThresholdPercentage, upperThresholdPercentage are in [0,1]. 
	 * For example: 
	 * lowerThresholdPercentage = 0.1 
	 * upperThresholdPercentage = 0.1 
	 * 
	 * This type of normalization is usefull when most of the image data is found in a small interval, 
	 * but there are outliers:        *  *
	 *                               **  * *
	 *                               *** * *      *
	 *                      *       ********      *
	 *                     ---------------------------
	 * For the above histogram we would like to stretch the main part and clamp the outliers.
	 *
	 */
	static void normalizeClamp(Image &img, 
		                         float lowerThresholdPercentage,
		                         float upperThresholdPercentage,
		                         float minVal=0.0f, float maxVal=255.0f);
	static void normalizeClamp(vector<Image *> &imgs, 
				                     float lowerThresholdPercentage,
				                     float upperThresholdPercentage,
				                     float minVal=0.0f, float maxVal=255.0f);
	/*
	 * Perfrom histogram equalization and then stretching to the interval [minVal,maxVal].
	 * Theory about histogram equalization can be found in:
	 *     "Digital Image Processing", Woods R.E. and Gonzalez R.C., 
	 *      Addison-Wesley, 1992, ISBN: 0201508036
         */
	static void histEqualize(Image &img, float minVal=0.0f, float maxVal=255.0);
	static void histEqualize(vector<Image *> &imgs, float minVal=0.0f, float maxVal=255.0);
private:	
	/*
	 * Histograms use the HistoRec struct as their data element.
	 * FloatComperator is responsible for the precision/epsilon which differntiates between
	 * two gray values.
	 */
	class FloatComparator {
	public:
	  FloatComparator(float precision) : PRECISION(precision) {}
	  bool operator()(const float val1, const float val2) const {return((val2-val1)>PRECISION);}
	private:
	  float PRECISION;
	};
  
  struct HistoRec {
    HistoRec(float val) : value(val), times(1) {}
    HistoRec(float val, unsigned tms) : value(val), times(tms) {}
    float value;
    unsigned times;
  };
  
  class HistoRecValComparator {
  public:
    HistoRecValComparator(FloatComparator *comparator) : _comparator(comparator) {}; 
    bool operator()(HistoRec *val1, HistoRec *val2) {return((*_comparator)(val1->value,val2->value));}
  private:
    FloatComparator *_comparator;
  };
  
  static void createHistogram(Image &img, vector<HistoRec *> &histogram);
	static void createHistogram(vector<Image *> &imgs, vector<HistoRec *> &histogram);
           //for debug purposes
  static void printHistogram(vector<HistoRec *> &histogram);
};


#endif //_HISTOGRAM_OP_
